package xbc.formula.engine.exp;

import java.math.RoundingMode;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import xbc.formula.engine.exp.datatype.XbcExpExpTree;
import xbc.formula.engine.exp.exception.XbcExpParamException;
import xbc.formula.engine.exp.exec.XbcExpCalculateExp;
import xbc.formula.engine.exp.parser.XbcExpParser;
import xbc.formula.engine.exp.vo.XbcExpRuleVo;
import xbc.formula.engine.exp.vo.XbcExpCalculatorReturnVo;
import xbc.formula.engine.exp.vo.XbcExpDrawnInfoVo;
import xbc.formula.engine.exp.vo.XbcExpResultVo;

/**
 * 表达式计算工具
 * 
 * @author leichangchun
 */
public class XbcExpCalculator {
	
	// 默认的大数计算保留小数位数
	private static final int DEFAULT_BAOLIU_WEISHU = 12;
	
	// 默认的取整模式: 四舍五入
	private static final RoundingMode DEFAULT_ROUND_MODE = RoundingMode.HALF_UP;

	/**
	 * 获取表达式里面的指标
	 * @param expStr 表达式字符串
	 * @return 指标列表
	 */
	public static List<String> getParamters(String expStr) {
		if(expStr == null) {
			throw new XbcExpParamException("请传入一个表达式。");
		}
		XbcExpParser ep = new XbcExpParser(expStr);
		ep.parse();
		Set<String> paramSet = ep.getParamSet();
		List<String> params = new ArrayList<>();
		for(String param : paramSet) {
			params.add(param);
		}
		return params;
	}

	public static List<XbcExpDrawnInfoVo> getDrawnInfoVos(String expStr) {
		if(expStr == null) {
			throw new XbcExpParamException("请传入一个表达式。");
		}
		XbcExpParser ep = new XbcExpParser(expStr, true);
		ep.parse();
		List<XbcExpDrawnInfoVo> divs = ep.getDrawnInfoVos();
		return divs;
	}
	
	/**
	 * 给定参数，计算表达式的值
	 * @param expStr 表达式字符串
	 * @param paramMap 指标Map
	 * @param formulaParams 表达式计算过程中，可以使用的参数(@XbcExpExtFunc)
	 * @return 表达式的计算结果
	 */
	public static Object calculate(String expStr, Map<String, Object> paramMap, Object[] formulaParams) {
		return calculate(DEFAULT_BAOLIU_WEISHU, DEFAULT_ROUND_MODE, expStr, paramMap, formulaParams);
	}
	
	public static Object calculate(String expStr, Map<String, Object> paramMap, List<XbcExpRuleVo> ruleVos, Object[] formulaParams) {
		return calculate(DEFAULT_BAOLIU_WEISHU, DEFAULT_ROUND_MODE, expStr, paramMap, ruleVos, formulaParams);
	}
	
	/**
	 * 给定参数，计算表达式的值
	 * @param weishu 保留的小数位数
	 * @param roundMode 保留小数时的取舍方式
	 * @param expStr 表达式字符串
	 * @param paramMap 指标Map
	 * @param formulaParams 表达式计算过程中，可以使用的参数(@XbcExpExtFunc)
	 * @return 表达式的计算结果
	 */
	public static Object calculate(int weishu, RoundingMode roundMode, String expStr, Map<String, Object> paramMap, Object[] formulaParams) {
		return calculate(weishu, roundMode, expStr, paramMap, null, formulaParams);
	}
	
	public static Object calculate(int weishu, RoundingMode roundMode, String expStr, Map<String, Object> paramMap, List<XbcExpRuleVo> ruleVos, Object[] formulaParams) {
		Map<String, XbcExpResultVo> piMap = new HashMap<>();
		return calculateExp(weishu, roundMode, expStr, paramMap, piMap, ruleVos, formulaParams);
	}
	
	public static XbcExpCalculatorReturnVo calculateEx(String expStr, Map<String, Object> paramMap, Object[] formulaParams) {
		return calculateEx(DEFAULT_BAOLIU_WEISHU, DEFAULT_ROUND_MODE, expStr, paramMap, formulaParams);
	}

	public static XbcExpCalculatorReturnVo calculateEx(int weishu, RoundingMode roundMode, String expStr, Map<String, Object> paramMap, Object[] formulaParams) {
		return calculateEx(weishu, roundMode, expStr, paramMap, null, formulaParams);
	}

	public static XbcExpCalculatorReturnVo calculateEx(int weishu, RoundingMode roundMode, String expStr, Map<String, Object> paramMap, List<XbcExpRuleVo> ruleVos, Object[] formulaParams) {
		Map<String, XbcExpResultVo> piMap = new HashMap<>();
		Object retObj = calculateExp(weishu, roundMode, expStr, paramMap, piMap, ruleVos, formulaParams);
		XbcExpCalculatorReturnVo crv = new XbcExpCalculatorReturnVo();
		crv.setResult(retObj);
		Map<String, Object> retParamMap = new HashMap<>();
		crv.setParamMap(retParamMap);
		for(String key : piMap.keySet()) {
			XbcExpResultVo result = piMap.get(key);
			retParamMap.put(key, result.value);
		}
		return crv;
	}

	private static Object calculateExp(int weishu, RoundingMode roundMode, String expStr, Map<String, Object> paramMap, Map<String, XbcExpResultVo> piMap, List<XbcExpRuleVo> ruleVos, Object[] formulaParams) {
		if(expStr == null) {
			throw new XbcExpParamException("请传入一个表达式。");
		}
		if(paramMap == null) {
			paramMap = new HashMap<>();
		}
		XbcExpParser ep = new XbcExpParser(expStr);
		XbcExpExpTree exp = ep.parse();
		Set<String> paramSet = ep.getParamSet();
		List<String> notExistedParams = new ArrayList<>();
		for(String param : paramSet) {
			if(!paramMap.containsKey(param)) {
				notExistedParams.add(param);
			}
		}
		if(notExistedParams.size() > 0) {
			String msg = String.join(",", notExistedParams);
			throw new XbcExpParamException("需要传递以下参数: " + msg + ".公式: " + expStr);
		}
		
		for(String key : paramMap.keySet()) {
			XbcExpResultVo result = new XbcExpResultVo();
			Object val = paramMap.get(key);
			result.value = val;
			piMap.put(key, result);
		}

		XbcExpCalculateExp calculateExp = new XbcExpCalculateExp(weishu, roundMode, expStr, piMap, ruleVos, formulaParams);
		Object ret = calculateExp.calculate(exp);
		return ret;
	}
}
