package xbc.formula.engine.exp.exec;

import java.math.BigDecimal;
import java.sql.Timestamp;
import java.util.Calendar;
import java.util.List;

import xbc.formula.engine.exp.datatype.XbcExpExpTree;
import xbc.formula.engine.exp.exception.XbcExpExecException;
import xbc.formula.engine.exp.util.XbcExpDateUtil;
import xbc.formula.engine.exp.vo.XbcExpCalculatorExpFuncDateTtcVo;
import xbc.formula.engine.exp.vo.XbcExpResultVo;

/**
 * 表达式计算引擎自带函数(日期类函数)执行类
 * 
 * @author leichangchun
 */
public class XbcExpCalculatorExpFuncDate {

	private XbcExpCalculatorExpFunc calculatorExpFunc;

	private String expStr;
	
	public XbcExpCalculatorExpFuncDate(XbcExpCalculatorExpFunc calculatorExpFunc, String expStr) {
		this.calculatorExpFunc = calculatorExpFunc;
		this.expStr = expStr;
	}

    protected void calcCompareDateFunc(XbcExpExpTree exp, XbcExpResultVo result) {
        if(exp.params.size() < 2) {
            throw new XbcExpExecException(exp.linenum + "行: compareDate函数需要两个日期类型的参数。公式: " + expStr);
        }
        List<Object> pvs = calculatorExpFunc.getFuncParams(exp);
        if(pvs.get(0) == null || pvs.get(1) == null) {
            throw new XbcExpExecException(exp.linenum + "行: compareDate函数的第一个参数和第二个参数都不能为空，且必须为日期类型的参数。公式: " + expStr);
        }
        XbcExpCalculatorExpFuncDateTtcVo ttcv1 = XbcExpDateUtil.translateToCalendar("compareDate", "一", pvs.get(0));
        XbcExpCalculatorExpFuncDateTtcVo ttcv2 = XbcExpDateUtil.translateToCalendar("compareDate", "二", pvs.get(1));
        int cr = ttcv1.getCv().compareTo(ttcv2.getCv());
        if(cr == 0) {
            result.value = 0;
        } else if(cr > 0) {
            result.value = 1;
        } else {
            result.value = -1;
        }
    }

    protected void calcAddDateFunc(XbcExpExpTree exp, XbcExpResultVo result) {
        calcCommonDateFunc("addDate", Calendar.DAY_OF_MONTH, exp, result);
    }

    protected void calcAddMonthFunc(XbcExpExpTree exp, XbcExpResultVo result) {
        calcCommonDateFunc("addMonth", Calendar.MONTH, exp, result);
    }

    protected void calcAddHourFunc(XbcExpExpTree exp, XbcExpResultVo result) {
        calcCommonDateFunc("addHour", Calendar.HOUR, exp, result);
    }

    protected void calcDateToStringFunc(XbcExpExpTree exp, XbcExpResultVo result) {
        if(exp.params.size() == 0) {
            throw new XbcExpExecException(exp.linenum + "行: dateToString函数需要一个或两个参数。公式: " + expStr);
        } else if(exp.params.size() > 2) {
            throw new XbcExpExecException(exp.linenum + "行: dateToString函数需要一个或两个参数。公式: " + expStr);
        }
        List<Object> pvs = calculatorExpFunc.getFuncParams(exp);
        XbcExpCalculatorExpFuncDateTtcVo ttcv = XbcExpDateUtil.translateToCalendar("dateToString", "一", pvs.get(0));
        String format = null;
        if(pvs.size() == 1 || pvs.get(1) == null) {
            format = "yyyy-MM-dd";
        } else {
            Object p2 = pvs.get(1);
            if(p2.getClass() != String.class) {
                throw new XbcExpExecException(exp.linenum + "行: dateToString函数的第二个参数应该是一个表示日期格式的字符串。公式: " + expStr);
            }
            format = (String)p2;
        }
        String ds = XbcExpDateUtil.calendarToString(ttcv.getCv(), format);
        result.value = ds;
    }

    protected void calcThisDayFunc(XbcExpExpTree exp, XbcExpResultVo result) {
        String ds = XbcExpDateUtil.thisDay();
        result.value = ds;
    }

    protected void calcNowFunc(XbcExpExpTree exp, XbcExpResultVo result) {
        String ds = XbcExpDateUtil.now();
        result.value = ds;
    }

    protected void calcDiffDateFunc(XbcExpExpTree exp, XbcExpResultVo result) {
        if(exp.params.size() < 2) {
            throw new XbcExpExecException(exp.linenum + "行: diffDate函数需要两个日期类型的参数。公式: " + expStr);
        }
        List<Object> pvs = calculatorExpFunc.getFuncParams(exp);
        if(pvs.get(0) == null || pvs.get(1) == null) {
            throw new XbcExpExecException(exp.linenum + "行: diffDate函数的第一个参数和第二个参数都不能为空，且必须为日期类型的参数。公式: " + expStr);
        }
        XbcExpCalculatorExpFuncDateTtcVo ttcv1 = XbcExpDateUtil.translateToCalendar("compareDate", "一", pvs.get(0));
        XbcExpCalculatorExpFuncDateTtcVo ttcv2 = XbcExpDateUtil.translateToCalendar("compareDate", "二", pvs.get(1));
        int iv = XbcExpDateUtil.diffDate(ttcv1.getCv(), ttcv2.getCv());
        result.value = iv;
    }

    protected void toDateFunc(XbcExpExpTree exp, XbcExpResultVo result) {
    	if(exp.params.size() != 1) {
            throw new XbcExpExecException(exp.linenum + "行: toDate函数需要一个日期/字符串类型的参数。");
        }
        List<Object> pvs = calculatorExpFunc.getFuncParams(exp);
        if(pvs.get(0) == null) {
            throw new XbcExpExecException(exp.linenum + "行: toDate函数的参数不能为空，且必须为日期/字符串类型的参数。");
        }
        Object pv = pvs.get(0);
        if(!XbcExpDateUtil.isDateType(pv)) {
            throw new XbcExpExecException(exp.linenum + "行: toDate函数的参数不是日期/字符串类型的参数， " + pv.getClass().getName() + "。");
        }
        XbcExpCalculatorExpFuncDateTtcVo iv = XbcExpDateUtil.translateToCalendar("toDate", "", pv);
        result.value = iv.getCv();
    }

    private void calcCommonDateFunc(String funcName, int field, XbcExpExpTree exp, XbcExpResultVo result) {
        if(exp.params.size() < 2) {
            throw new XbcExpExecException(exp.linenum + "行: " + funcName + "函数需要两个参数。公式: " + expStr);
        }
        List<Object> pvs = calculatorExpFunc.getFuncParams(exp);
        if(pvs.get(0) == null || pvs.get(1) == null) {
            throw new XbcExpExecException(exp.linenum + "行: " + funcName + "函数的第一个参数和第二个参数都不能为空。公式: " + expStr);
        }
        XbcExpCalculatorExpFuncDateTtcVo ttcv = XbcExpDateUtil.translateToCalendar(funcName, "一", pvs.get(0));
        int oi = translateToInt(funcName, "二", pvs.get(1), exp.linenum);
        Calendar cv = ttcv.getCv();
        cv.add(field, oi);
        Class<?> clazz = ttcv.getClazz();
        Object retOv = null;
        if(clazz == java.util.Date.class) {
            java.util.Date dv = new java.util.Date();
            dv.setTime(cv.getTimeInMillis());
            retOv = dv;
        } else if(clazz == java.sql.Date.class) {
            java.sql.Date dv = new java.sql.Date(cv.getTimeInMillis());
            retOv = dv;
        } else if(clazz == Timestamp.class) {
            Timestamp tv = new Timestamp(cv.getTimeInMillis());
            retOv = tv;
        } else if(Calendar.class.isAssignableFrom(clazz)) {
            retOv = cv;
        }
        result.value = retOv;
    }

    private int translateToInt(String funcName, String pidx, Object ov, int linenum) {
        Class<?> clazz = ov.getClass();
        if(clazz == int.class || clazz == Integer.class) {
            int iv = (int)ov;
            return iv;
        }
        if(clazz == long.class || clazz == Long.class) {
            long lv = (long)ov;
            return (int)lv;
        }
        if(clazz == BigDecimal.class) {
            BigDecimal bv = (BigDecimal)ov;
            return bv.intValue();
        }
        throw new XbcExpExecException(linenum + "行: " + funcName + "函数的第" + pidx + "个参数必须为整型的参数。公式: " + expStr);
    }
}
