package xbc.formula.engine.exp.exec;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.sql.Date;
import java.sql.Timestamp;
import java.util.List;

import xbc.formula.engine.exp.datatype.XbcExpExpTree;
import xbc.formula.engine.exp.datatype.XbcExpToken;
import xbc.formula.engine.exp.exception.XbcExpExecException;
import xbc.formula.engine.exp.util.XbcExpDateUtil;
import xbc.formula.engine.exp.util.XbcExpNumericUtil;
import xbc.formula.engine.exp.vo.XbcExpCalculatorExpFuncDateTtcVo;
import xbc.formula.engine.exp.vo.XbcExpIntervalVo;
import xbc.formula.engine.exp.vo.XbcExpResultVo;

/**
 * 层差表达式执行类
 * 
 * 层差表达式, 例如：
 *    exp { >= 0 : 1, (0, -10] : 0.9, (-10, -20] : 0.8, ...., < 100 : 0}
 *    即: 当 exp 的值落在不同的区间时，取不同的值
 *    
 * @author leichangchun
 */
public class XbcExpCalculatorExpItv {

	private XbcExpCalculateExp calculateExp;

	private int weishu;
	
	private RoundingMode roundMode;

	private String expStr;
	
	public XbcExpCalculatorExpItv(XbcExpCalculateExp calculateExp, int weishu, RoundingMode roundMode, String expStr) {
		this.calculateExp = calculateExp;
		this.weishu = weishu;
		this.roundMode = roundMode;
		this.expStr = expStr;
	}

	/**
	 * 层差表达式计算
	 * 
	 * 2022/10/29 不再要求指标表达式的结果一定是数值型，除数值类型外，允许字符串类型和日期、时间类型作为指标值
	 * 
	 * @param exp
	 * @param result
	 */
    public void calcItvExp(XbcExpExpTree exp, XbcExpResultVo result) {
        XbcExpResultVo idr = new XbcExpResultVo();
        calculateExp.calcExp(exp.lchild, idr);
        if(idr.value == null) {
            if(exp.rchild == null) {
                throw new XbcExpExecException(exp.linenum + "行: 层差表达式的指标值为null，且没有设置默认值。");
            }
            // 当指标值为空时，返回默认值
            calculateExp.calcExp(exp.rchild, result);
            return;
        }
        List<XbcExpIntervalVo> itvs = exp.itvs;
        if(itvs == null || itvs.size() == 0) {
            throw new XbcExpExecException(exp.linenum + "行: 层差计算没有设置区间。");
        }
        for(int k = 0;k < itvs.size();k ++) {
            boolean matched = false;
            XbcExpIntervalVo itv = itvs.get(k);
            if(itv.token == XbcExpToken.LT_TT || itv.token == XbcExpToken.LEQ_TT || itv.token == XbcExpToken.GT_TT || itv.token == XbcExpToken.GEQ_TT || itv.token == XbcExpToken.EQ_TT) {
                XbcExpResultVo sr = new XbcExpResultVo();
                calculateExp.calcExp(itv.startExp, sr);
                if(sr.value != null) {
                	int cr = compareI2Iv(idr.value, sr.value, k + 1, false, false, exp.linenum);
                    if(cr < 0 && itv.token == XbcExpToken.LT_TT) {
                        matched = true;
                    } else if(cr <= 0 && itv.token == XbcExpToken.LEQ_TT) {
                        matched = true;
                    } else if(cr > 0 && itv.token == XbcExpToken.GT_TT) {
                        matched = true;
                    } else if(cr >= 0 && itv.token == XbcExpToken.GEQ_TT) {
                        matched = true;
                    } else if(cr == 0 && itv.token == XbcExpToken.EQ_TT) {
                        matched = true;
                    }
                }
            } else if(itv.startExp == null || itv.endExp == null) {
                throw new XbcExpExecException(exp.linenum + "行: 层差表达式必须同时设置起点和终点");
            } else {
                XbcExpResultVo sr = new XbcExpResultVo();
                calculateExp.calcExp(itv.startExp, sr);
                XbcExpResultVo er = new XbcExpResultVo();
                calculateExp.calcExp(itv.endExp, er);
                if(sr.value != null && er.value != null) {
                	int scr = compareI2Iv(idr.value, sr.value, k + 1, true, false, exp.linenum);
                    int ecr = compareI2Iv(idr.value, er.value, k + 1, false, false, exp.linenum);
                    if(compareI2Iv(sr.value, er.value, k + 1, true, true, exp.linenum) <= 0) { // 正向区间
                        if(itv.startType == XbcExpIntervalVo.OPEN_INTERVAL && itv.endType == XbcExpIntervalVo.OPEN_INTERVAL) {
                            if(scr > 0 && ecr < 0) {
                                matched = true;
                            }
                        } else if(itv.startType == XbcExpIntervalVo.OPEN_INTERVAL && itv.endType == XbcExpIntervalVo.CLOSE_INTERVAL) {
                            if(scr > 0 && ecr <= 0) {
                                matched = true;
                            }
                        } else if(itv.startType == XbcExpIntervalVo.CLOSE_INTERVAL && itv.endType == XbcExpIntervalVo.OPEN_INTERVAL) {
                            if(scr >= 0 && ecr < 0) {
                                matched = true;
                            }
                        } else {
                            if(scr >= 0 && ecr <= 0) {
                                matched = true;
                            }
                        }
                    } else { // 反向区间
                        if(itv.startType == XbcExpIntervalVo.OPEN_INTERVAL && itv.endType == XbcExpIntervalVo.OPEN_INTERVAL) {
                            if(scr < 0 && ecr > 0) {
                                matched = true;
                            }
                        } else if(itv.startType == XbcExpIntervalVo.OPEN_INTERVAL && itv.endType == XbcExpIntervalVo.CLOSE_INTERVAL) {
                            if(scr < 0 && ecr >= 0) {
                                matched = true;
                            }
                        } else if(itv.startType == XbcExpIntervalVo.CLOSE_INTERVAL && itv.endType == XbcExpIntervalVo.OPEN_INTERVAL) {
                            if(scr <= 0 && ecr > 0) {
                                matched = true;
                            }
                        } else {
                            if(scr <= 0 && ecr >= 0) {
                                matched = true;
                            }
                        }
                    }
                }
            }
            if(matched) {
            	calculateExp.calcExp(itv.valueExp, result);
                return;
            }
        }
        if(exp.rchild != null) {
            // 如果没有落到任何区间，且指定了默认值，则使用默认值
        	calculateExp.calcExp(exp.rchild, result);
            return;
        }
        // 没有落到任何区间，且没有指定默认值，则返回 0
        result.value = 0;
    }

    private static boolean isNumeric(Object ov) {
        Class<?> c = ov.getClass();
        if(c == int.class || c == long.class || c == byte.class || c == short.class || c == double.class || c == float.class) {
            return true;
        }
        if(c == BigDecimal.class || c == Integer.class || c == Long.class || c == Byte.class || c == Short.class || c == Double.class || c == Float.class) {
            return true;
        }
        return false;
    }

    private int compareI2Iv(Object iv, Object iiv, int index, boolean isS, boolean isI, int linenum) {
        if(isNumeric(iv) && isNumeric(iiv)) {
            // 两个数值进行比较
            BigDecimal bv1 = XbcExpNumericUtil.translate2BigDecimal(iv, weishu, roundMode, expStr);
            BigDecimal bv2 = XbcExpNumericUtil.translate2BigDecimal(iiv, weishu, roundMode, expStr);
            int rc = bv1.compareTo(bv2);
            return rc;
        }
        if(XbcExpDateUtil.isDateType(iv) && XbcExpDateUtil.isDateType(iiv)) {
            // 两个日期类型的数值进行比较
            XbcExpCalculatorExpFuncDateTtcVo tcvl = XbcExpDateUtil.translateToCalendar("", "", iv);
            XbcExpCalculatorExpFuncDateTtcVo tcvr = XbcExpDateUtil.translateToCalendar("", "", iiv);
            int rc = tcvl.getCv().compareTo(tcvr.getCv());
            return rc;
        }
        if(iv.getClass() == String.class && iiv.getClass() == String.class) {
            // 两个字符串进行比较
            String sv1 = (String)iv;
            String sv2 = (String)iiv;
            int rc = sv1.compareTo(sv2);
            return rc;
        }
        if(XbcExpDateUtil.isDateType(iv) && iiv.getClass() == String.class) {
            Date dv = XbcExpDateUtil.dateFromString((String)iiv);
            if(dv != null) {
                XbcExpCalculatorExpFuncDateTtcVo tcvl = XbcExpDateUtil.translateToCalendar("", "", iv);
                XbcExpCalculatorExpFuncDateTtcVo tcvr = XbcExpDateUtil.translateToCalendar("", "", dv);
                int rc = tcvl.getCv().compareTo(tcvr.getCv());
                return rc;
            } else {
                Timestamp tv = XbcExpDateUtil.timestampFromString((String)iiv);
                if(tv != null) {
                    XbcExpCalculatorExpFuncDateTtcVo tcvl = XbcExpDateUtil.translateToCalendar("", "", iv);
                    XbcExpCalculatorExpFuncDateTtcVo tcvr = XbcExpDateUtil.translateToCalendar("", "", tv);
                    int rc = tcvl.getCv().compareTo(tcvr.getCv());
                    return rc;
                }
            }
        }
        if(iv.getClass() == String.class && XbcExpDateUtil.isDateType(iiv)) {
            Date dv = XbcExpDateUtil.dateFromString((String)iv);
            if(dv != null) {
                XbcExpCalculatorExpFuncDateTtcVo tcvl = XbcExpDateUtil.translateToCalendar("", "", dv);
                XbcExpCalculatorExpFuncDateTtcVo tcvr = XbcExpDateUtil.translateToCalendar("", "", iiv);
                int rc = tcvl.getCv().compareTo(tcvr.getCv());
                return rc;
            } else {
                Timestamp tv = XbcExpDateUtil.timestampFromString((String)iv);
                if(tv != null) {
                    XbcExpCalculatorExpFuncDateTtcVo tcvl = XbcExpDateUtil.translateToCalendar("", "", tv);
                    XbcExpCalculatorExpFuncDateTtcVo tcvr = XbcExpDateUtil.translateToCalendar("", "", iiv);
                    int rc = tcvl.getCv().compareTo(tcvr.getCv());
                    return rc;
                }
            }
        }
        if(isI) {
            String err = linenum + "行: 层差计算， 第" + index + "区间的起点值类型【" + iv.getClass().getName() + "】与终点值类型【" + iiv.getClass().getName() + "】无法进行比较。";
            throw new XbcExpExecException(err);
        } else {
            String err = linenum + "行: 层差计算， 第" + index + "区间的";
            if(isS) {
                err = err + "起点值";
            } else {
                err = err + "终点值";
            }
            err = err + "的类型【" + iiv.getClass().getName() + "】无法与指标值类型【" + iv.getClass().getName() + "】进行比较。";
            throw new XbcExpExecException(err);
        }
    }
}
