package xbc.formula.engine.exp.parser;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import xbc.formula.engine.exp.datatype.XbcExpExpKind;
import xbc.formula.engine.exp.datatype.XbcExpExpOpKind;
import xbc.formula.engine.exp.datatype.XbcExpExpTree;
import xbc.formula.engine.exp.datatype.XbcExpExpVarKind;
import xbc.formula.engine.exp.datatype.XbcExpToken;
import xbc.formula.engine.exp.exception.XbcExpParseException;
import xbc.formula.engine.exp.scanner.XbcExpScanner;
import xbc.formula.engine.exp.vo.XbcExpDrawnInfoVo;
import xbc.formula.engine.exp.vo.XbcExpIntervalVo;

/**
 * 语法分析，最总得到表达式树
 * 
 * @author leichangchun
 */
public class XbcExpParser {

	private String expStr;
	
	private XbcExpScanner es;
	
	private int token;
	
	private Set<String> paramSet;
	
	private Set<String> removedParamSet;

	private boolean mkDiv;

	public XbcExpParser(String expStr) {
		this(expStr, false);
	}
	public XbcExpParser(String expStr, boolean mkDiv) {
		this.expStr = expStr;
		this.mkDiv = mkDiv;
		paramSet = new HashSet<>();
		removedParamSet = new HashSet<>();
	}
	
	public Set<String> getParamSet() {
		return paramSet;
	}
	
	public List<XbcExpDrawnInfoVo> getDrawnInfoVos() {
		if(es == null) {
			return null;
		}
		return es.getDrawnInfoVos();
	}
	
	public XbcExpExpTree parse() {
		es = new XbcExpScanner(expStr, mkDiv);
		token = es.readToken();
		XbcExpExpTree exp = readExpSequence();
		return exp;
	}

	private XbcExpExpTree readExpSequence() {
		XbcExpExpTree exp = readNextExp();
		XbcExpExpTree next = exp;
		while(next != null && next.next != null) {
			next = next.next;
		}
		while(next != null) {
			next.next = readNextExp();
			if(next.next == null) {
				break;
			}
			while(next.next != null) {
				next = next.next;
			}
		}
		return exp;
	}

	private XbcExpExpTree readNextExp() {
		XbcExpExpTree exp = null;
		switch(token) {
		case XbcExpToken.EXPNAME_TT: // variable-name
		case XbcExpToken.INT_TT:     // integer
		case XbcExpToken.DBL_TT:     // double
		case XbcExpToken.STR_TT:     // string
		case XbcExpToken.LPRN_TT:    // (
		case XbcExpToken.SELFMNS_TT: // --
		case XbcExpToken.SELFPLS_TT: // ++
		case XbcExpToken.PLS_TT:     // +
		case XbcExpToken.MNS_TT:     // -
		case XbcExpToken.BNOT_TT:    // ~
		case XbcExpToken.NOT_TT:     // !
		case XbcExpToken.NEW_TT:     // new
		case XbcExpToken.INVOKE_TT:
			exp = readAsnExp();
			break;
		case XbcExpToken.FOR_TT:     // for
			exp = readForExp();
			break;
		case XbcExpToken.BREAK_TT:   // break;
			exp = readBreakExp();
			break;
		case XbcExpToken.CONTINUE_TT: // continue
			exp = readContinueExp();
			break;
		case XbcExpToken.WHILE_TT:    // while
			exp = readWhileExp();
			break;
		case XbcExpToken.IF_TT:      // if
			exp = readIfExp();
			break;
		case XbcExpToken.LSQB_TT:    // {
			exp = readBlockExp();
			break;
		case XbcExpToken.DIV0_TT:     // div0
			exp = readDiv0Exp();
			break;
		case XbcExpToken.EXIT_TT:     // exit
			exp = readExitExp();
			break;
		case XbcExpToken.AT_TT:     // exit
			exp = readAtExp();
			break;
		case XbcExpToken.SEMI_TT:
			match(XbcExpToken.SEMI_TT);
			exp = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.NULL_OK);
			break;
		case XbcExpToken.EOF_TT:
			break;
		default:
			throw new XbcExpParseException(es.getLinenum() + "行: 未知的符号: " + XbcExpToken.getTokenName(token) + ", 公式: " + expStr);
		}
		return exp;
	}
	
	private void match(int expectedToken) {
		if(expectedToken != token) {
			throw new XbcExpParseException(es.getLinenum() + "行: 希望符号【" + XbcExpToken.getTokenName(expectedToken) + "】, 实际符号【" + XbcExpToken.getTokenName(token) + "】。" + " 公式: " + expStr);
		}
		token = es.readToken();
	}

	private XbcExpExpTree readAtExp() {
		match(XbcExpToken.AT_TT);
		if(token != XbcExpToken.EXPNAME_TT) {
			throw new XbcExpParseException(es.getLinenum() + "行: @后面必须接【Rule】, 公式: " + expStr);
		}
		String tokenContent = es.getTokenContent();
		if(!tokenContent.equals("Rule")) {
			throw new XbcExpParseException(es.getLinenum() + "行: @后面必须接【Rule】, 公式: " + expStr);
		}
		match(XbcExpToken.EXPNAME_TT);
		match(XbcExpToken.LPRN_TT);
		XbcExpExpTree exp = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.AT_OK);
		List<XbcExpExpTree> params = new ArrayList<>();
		exp.params = params;
		while(token != XbcExpToken.RPRN_TT) {
			if(token == XbcExpToken.SEMI_TT) {
				match(token);
				continue;
			}
			if(token == XbcExpToken.CMM_TT) {
				match(token);
			}
			if(token != XbcExpToken.EXPNAME_TT) {
				throw new XbcExpParseException(es.getLinenum() + "行: @Rule里面只能定义属性 type, name, displayOrder, bizIdCode, desc, comment, params, record, notrecord, 公式: " + expStr);
			}
			String tc = es.getTokenContent();
			if(!tc.equals("type") && !tc.equals("name") && !tc.equals("desc") && !tc.equals("comment") && !tc.equals("displayOrder") && !tc.equals("bizIdCode") && !tc.equals("params") && !tc.equals("record") && !tc.equals("notrecord")) {
				throw new XbcExpParseException(es.getLinenum() + "行: @Rule里面只能定义属性 type, name, displayOrder, bizIdCode, desc, comment params, record, notrecord, 公式: " + expStr);
			}
			match(token);
			match(XbcExpToken.ASN_TT);
			if(tc.equals("params")) {
				if(exp.ruleParams != null) {
					throw new XbcExpParseException(es.getLinenum() + "行: @Rule里面重复定义了属性: " + tc + ", 公式: " + expStr);
				}
				List<XbcExpExpTree> rps = new ArrayList<>();
				exp.ruleParams = rps;
				match(XbcExpToken.LSQB_TT);
				while(token != XbcExpToken.RSQB_TT) {
					if(token == XbcExpToken.CMM_TT) {
						match(token);
					}
					if(token != XbcExpToken.EXPNAME_TT) {
						throw new XbcExpParseException(es.getLinenum() + "行: params里面只能填写变量名, 公式: " + expStr);
					}
					XbcExpExpTree rp = XbcExpExpTree.newIdExptree(es.getLinenum(), es.getTokenContent());
					rps.add(rp);
					match(token);
				}
				match(token);
			} else {
				for(XbcExpExpTree pe : params) {
					String sv = (String)(pe.ruleName);
					if(sv.equals(tc)) {
						throw new XbcExpParseException(es.getLinenum() + "行: @Rule里面重复定义了属性: " + tc + ", 公式: " + expStr);
					}
				}
				XbcExpExpTree pe = readAsnExp();
				pe.ruleName = tc;
				params.add(pe);
			}
		}
		match(token);
		exp.lchild = readNextExp();
		return exp;
	}

	private XbcExpExpTree readExitExp() {
		match(XbcExpToken.EXIT_TT);
		XbcExpExpTree exp = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.EXIT_OK);
		exp.lchild = readAsnExp();
		return exp;
	}

	private XbcExpExpTree readInvokeExp() {
		String tokenContent = es.getTokenContent();
		match(XbcExpToken.INVOKE_TT);
		XbcExpExpTree exp = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.INVOKE_OK);
		exp.value = tokenContent;
		exp.params = new ArrayList<>();

		match(XbcExpToken.LPRN_TT);
		while(token != XbcExpToken.RPRN_TT) {
			XbcExpExpTree pexp = readAsnExp();
			exp.params.add(pexp);
			if(token == XbcExpToken.CMM_TT) {
				match(token);
				if(token == XbcExpToken.RPRN_TT) {
					throw new XbcExpParseException(es.getLinenum() + "行: 解析函数【" + tokenContent + "】的参数时出错。" + " 公式: " + expStr);
				}
			} else if(token != XbcExpToken.RPRN_TT) {
				throw new XbcExpParseException(es.getLinenum() + "行: 解析函数【" + tokenContent + "】的参数时出错。" + " 公式: " + expStr);
			}
		}
		match(XbcExpToken.RPRN_TT);
		
		return exp;
	}
	
	private XbcExpExpTree readNewExp() {
		String tokenContent = es.getTokenContent();
		match(XbcExpToken.NEW_TT);
		XbcExpExpTree exp = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.NEW_OK);
		exp.value = tokenContent;
		exp.params = new ArrayList<>();

		match(XbcExpToken.LPRN_TT);
		while(token != XbcExpToken.RPRN_TT) {
			XbcExpExpTree pexp = readAsnExp();
			exp.params.add(pexp);
			if(token == XbcExpToken.CMM_TT) {
				match(token);
				if(token == XbcExpToken.RPRN_TT) {
					throw new XbcExpParseException(es.getLinenum() + "行: 解析函数【" + tokenContent + "】的参数时出错。" + " 公式: " + expStr);
				}
			} else if(token != XbcExpToken.RPRN_TT) {
				throw new XbcExpParseException(es.getLinenum() + "行: 解析函数【" + tokenContent + "】的参数时出错。" + " 公式: " + expStr);
			}
		}
		match(XbcExpToken.RPRN_TT);
		
		return exp;
	}

	private XbcExpExpTree readDiv0Exp() {
		match(XbcExpToken.DIV0_TT);
		XbcExpExpTree exp = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.DIV0_OK);
		match(XbcExpToken.LPRN_TT);
		XbcExpExpTree texp = readAsnExp();
		match(XbcExpToken.RPRN_TT);
		exp.lchild = texp;
		return exp;
	}

	private XbcExpExpTree readBlockExp() {
		match(XbcExpToken.LSQB_TT);
		XbcExpExpTree exp = null;
		if(token == XbcExpToken.RSQB_TT) {
			match(XbcExpToken.RSQB_TT);
			exp = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.NULL_OK);
		} else {
			exp = readNextExp();
			XbcExpExpTree tail = exp;
			XbcExpExpTree next = exp;
			while(tail.next != null) {
				tail = tail.next;
			}
			while(token != XbcExpToken.RSQB_TT) {
				if(next == null) {
					throw new XbcExpParseException(es.getLinenum() + "行: 没有结束的\"}\"。 公式: " + expStr);
				}
				next = readNextExp();
				tail.next = next;
				tail = next;
				if(next != null) {
					while(tail.next != null) {
						tail = tail.next;
					}
				}
			}
			match(XbcExpToken.RSQB_TT);
		}
		return exp;
	}
	
	private XbcExpExpTree readIfExp() {
		match(XbcExpToken.IF_TT);
		XbcExpExpTree exp = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.IF_OK);
		match(XbcExpToken.LPRN_TT);
		exp.lchild = readAsnExp();
		match(XbcExpToken.RPRN_TT);
		exp.mchild = readNextExp();
		if(token == XbcExpToken.ELSE_TT) {
			match(XbcExpToken.ELSE_TT);
			exp.rchild = readNextExp();
		}
		return exp;
	}

	private XbcExpExpTree readBreakExp() {
		match(XbcExpToken.BREAK_TT);
		XbcExpExpTree exp = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.BREAK_OK);
		return exp;
	}

	private XbcExpExpTree readContinueExp() {
		match(XbcExpToken.CONTINUE_TT);
		XbcExpExpTree exp = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.CONTINUE_OK);
		return exp;
	}
	
	private XbcExpExpTree readWhileExp() {
		match(XbcExpToken.WHILE_TT);
		XbcExpExpTree exp = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.WHILE_OK);
		match(XbcExpToken.LPRN_TT);
		exp.lchild = readAsnExp();
		match(XbcExpToken.RPRN_TT);
		exp.body = readNextExp();
		return exp;
	}
	
	private XbcExpExpTree readForExp() {
		match(XbcExpToken.FOR_TT);
		XbcExpExpTree exp = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.FOR_OK);
		match(XbcExpToken.LPRN_TT);
		if(token != XbcExpToken.SEMI_TT) {
			exp.lchild = readAsnExp();
		} else {
			exp.lchild = null;
		}
		match(XbcExpToken.SEMI_TT);
		if(token != XbcExpToken.SEMI_TT) {
			exp.mchild = readAsnExp();
		} else {
			exp.mchild = null;
		}
		match(XbcExpToken.SEMI_TT);
		if(token != XbcExpToken.SEMI_TT) {
			exp.rchild = readAsnExp();
		} else {
			exp.rchild = null;
		}
		match(XbcExpToken.RPRN_TT);
		exp.body = readNextExp();
		return exp;
	}
	
	private int translateAsnEt(int t) {
		switch(t) {
	    case XbcExpToken.MODASN_TT:
	        return XbcExpExpOpKind.MODASN_OK;
	    case XbcExpToken.BANDASN_TT:
	        return XbcExpExpOpKind.BANDASN_OK;
	    case XbcExpToken.MPTASN_TT:
	        return XbcExpExpOpKind.MPTASN_OK;
	    case XbcExpToken.DIVASN_TT:
	        return XbcExpExpOpKind.DIVASN_OK;
	    case XbcExpToken.BXORASN_TT:
	        return XbcExpExpOpKind.BXORASN_OK;
	    case XbcExpToken.BORASN_TT:
	        return XbcExpExpOpKind.BORASN_OK;
	    case XbcExpToken.PLSASN_TT:
	        return XbcExpExpOpKind.PLSASN_OK;
	    case XbcExpToken.S2LASN_TT:
	        return XbcExpExpOpKind.S2LASN_OK;
	    case XbcExpToken.ASN_TT:
	        return XbcExpExpOpKind.ASN_OK;
	    case XbcExpToken.MNSASN_TT:
	        return XbcExpExpOpKind.MNSASN_OK;
	    case XbcExpToken.S2RASN_TT:
	        return XbcExpExpOpKind.S2RASN_OK;
		}
		return 0;
	}
	// 最基础的表达式
	private XbcExpExpTree readAsnExp() {
		XbcExpExpTree exp = readCndExp();
		int op = translateAsnEt(token);
		if(op > 0) {
			if(token == XbcExpToken.ASN_TT && exp.ek == XbcExpExpKind.ID_EK) {
				paramSet.remove(exp.value); // 如果是一个用等于符号的赋值语句，则把赋值语句前的变量从要求的参数里面去掉
				removedParamSet.add((String)exp.value);
			}
			XbcExpExpTree exp1 = new XbcExpExpTree(es.getLinenum(), op);
			match(token);
			exp1.lchild = exp;
			exp1.rchild = readAsnExp();
			exp = exp1;
		}
		// 读取层差表达式
		if(token == XbcExpToken.LSQB_TT) {
			XbcExpExpTree itvExp = readItvExp(exp);
			exp = itvExp;
		}
		return exp;
	}

	private XbcExpExpTree readItvExp(XbcExpExpTree lchild) {
		match(XbcExpToken.LSQB_TT);
		XbcExpExpTree exp = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.ITV_OK);
		exp.lchild = lchild;
		List<XbcExpIntervalVo> itvs = new ArrayList<>();
		exp.itvs = itvs;
		while(true) {   // 读取区间列表
			if(token == XbcExpToken.CLN_TT) {
				if(exp.rchild != null) {
					throw new XbcExpParseException(es.getLinenum() + "行: 层差表达式里面，只能设置一个默认值。 公式: " + expStr);
				}
				match(token);
				exp.rchild = readAsnExp();
				if(token == XbcExpToken.RSQB_TT) {
					match(token);
					break;
				}
				match(XbcExpToken.CMM_TT);
			}
			XbcExpIntervalVo itv = new XbcExpIntervalVo();
			itvs.add(itv);
			if(token == XbcExpToken.LT_TT || token == XbcExpToken.LEQ_TT || token == XbcExpToken.GT_TT || token == XbcExpToken.GEQ_TT || token == XbcExpToken.EQ_TT) {
				itv.token = token;
				match(token);
				itv.startExp = readAsnExp();
			} else {
				if(token == XbcExpToken.LPRN_TT) {  // ( 开区间
					itv.startType = XbcExpIntervalVo.OPEN_INTERVAL;
				} else if(token == XbcExpToken.LBKT_TT) {  // [ 闭区间
					itv.startType = XbcExpIntervalVo.CLOSE_INTERVAL;
				} else {  // , 无起点区间
					throw new XbcExpParseException(es.getLinenum() + "行: 希望符号【" + XbcExpToken.getTokenName(XbcExpToken.LPRN_TT) + "】或【" +
							XbcExpToken.getTokenName(XbcExpToken.LBKT_TT) + "】, 实际符号【" + XbcExpToken.getTokenName(token) + "】。" + " 公式: " + expStr);
				}
				match(token);
				if(token == XbcExpToken.CMM_TT) {
					itv.startExp = null;
				} else {
					itv.startExp = readAsnExp();
				}
				match(XbcExpToken.CMM_TT);
				if(token == XbcExpToken.RPRN_TT || token == XbcExpToken.RBKT_TT) {
					if(itv.startExp == null) {
						throw new XbcExpParseException(es.getLinenum() + "行: 层差表达式的区间，不能起点和终点都为空。 公式: " + expStr);
					}
					itv.endExp = null;
				} else {
					itv.endExp = readAsnExp();
				}
				if(token == XbcExpToken.RPRN_TT) {
					itv.endType = XbcExpIntervalVo.OPEN_INTERVAL;
				} else if(token == XbcExpToken.RBKT_TT) {
					itv.endType = XbcExpIntervalVo.CLOSE_INTERVAL;
				} else {
					throw new XbcExpParseException(es.getLinenum() + "行: 希望符号【" + XbcExpToken.getTokenName(XbcExpToken.RPRN_TT) + "】或【" +
							XbcExpToken.getTokenName(XbcExpToken.RBKT_TT) + "】, 实际符号【" + XbcExpToken.getTokenName(token) + "】。" + " 公式: " + expStr);
				}
				match(token);
			}
			match(XbcExpToken.CLN_TT);
			itv.valueExp = readAsnExp();
			if(token == XbcExpToken.RSQB_TT) {
				match(token);
				break;
			}
			if(token == XbcExpToken.CMM_TT) {
				match(token);
			} else {
				throw new XbcExpParseException(es.getLinenum() + "行: 希望符号【" + XbcExpToken.getTokenName(XbcExpToken.RSQB_TT) + "】或【" +
						XbcExpToken.getTokenName(XbcExpToken.CMM_TT) + "】, 实际符号【" + XbcExpToken.getTokenName(token) + "】。" + " 公式: " + expStr);
			}
		}
		return exp;
	}

	private XbcExpExpTree readCndExp() {
		XbcExpExpTree exp = readOrExp();
		if(token == XbcExpToken.QRY_TT) {
			match(token);
			XbcExpExpTree exp1 = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.QRY_OK);
			exp1.lchild = exp;
			exp1.rchild = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.NULL_OK);
			exp = exp1;
			exp1 = exp1.rchild;
			exp1.lchild = readAsnExp();
			match(XbcExpToken.CLN_TT);
			exp1.rchild = readAsnExp();
		}
		return exp;
	}

	private XbcExpExpTree readOrExp() {
		XbcExpExpTree exp = readAndExp();
		while(token == XbcExpToken.OR_TT) {
			match(token);
			XbcExpExpTree exp1 = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.OR_OK);
			exp1.lchild = exp;
			exp1.rchild = readAndExp();
			exp = exp1;
		}
		return exp;
	}

	private XbcExpExpTree readAndExp() {
		XbcExpExpTree exp = readBorExp();
		while(token == XbcExpToken.AND_TT) {
			match(token);
			XbcExpExpTree exp1 = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.AND_OK);
			exp1.lchild = exp;
			exp1.rchild = readBorExp();
			exp = exp1;
		}
		return exp;
	}

	private XbcExpExpTree readBorExp() {
		XbcExpExpTree exp = readBxorExp();
		while(token == XbcExpToken.BOR_TT) {
			match(token);
			XbcExpExpTree exp1 = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.BOR_OK);
			exp1.lchild = exp;
			exp1.rchild = readBxorExp();
			exp = exp1;
		}
		return exp;
	}

	private XbcExpExpTree readBxorExp() {
		XbcExpExpTree exp = readBandExp();
		while(token == XbcExpToken.BXOR_TT) {
			match(token);
			XbcExpExpTree exp1 = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.BXOR_OK);
			exp1.lchild = exp;
			exp1.rchild = readBandExp();
			exp = exp1;
		}
		return exp;
	}

	private XbcExpExpTree readBandExp() {
		XbcExpExpTree exp = readEquExp();
		while(token == XbcExpToken.BAND_TT) {
			match(token);
			XbcExpExpTree exp1 = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.BAND_OK);
			exp1.lchild = exp;
			exp1.rchild = readEquExp();
			exp = exp1;
		}
		return exp;
	}

	private int translateEquEt(int t) {
		switch(t) {
	    case XbcExpToken.EQ_TT:
	        return XbcExpExpOpKind.EQ_OK;
	    case XbcExpToken.NEQ_TT:
	        return XbcExpExpOpKind.NEQ_OK;
		}
		return 0;
	}
	
	private XbcExpExpTree readEquExp() {
		XbcExpExpTree exp = readRelExp();
		int op = translateEquEt(token);
		while(op > 0) {
			match(token);
			XbcExpExpTree exp1 = new XbcExpExpTree(es.getLinenum(), op);
			exp1.lchild = exp;
			exp1.rchild = readRelExp();
			exp = exp1;
			op = translateEquEt(token);
		}
		return exp;
	}

	private int translateRelEt(int t) {
		switch(t) {
	    case XbcExpToken.LT_TT:
	        return XbcExpExpOpKind.LT_OK;
	    case XbcExpToken.LEQ_TT:
	        return XbcExpExpOpKind.LEQ_OK;
	    case XbcExpToken.GT_TT:
	        return XbcExpExpOpKind.GT_OK;
	    case XbcExpToken.GEQ_TT:
	        return XbcExpExpOpKind.GEQ_OK;
		}
		return 0;
	}
	
	private XbcExpExpTree readRelExp() {
		XbcExpExpTree exp = readShiftExp();
		int op = translateRelEt(token);
		while(op > 0) {
			match(token);
			XbcExpExpTree exp1 = new XbcExpExpTree(es.getLinenum(), op);
			exp1.lchild = exp;
			exp1.rchild = readShiftExp();
			exp = exp1;
			op = translateRelEt(token);
		}
		return exp;
	}

	private int translateShiftEt(int t) {
		switch(t) {
	    case XbcExpToken.S2L_TT:
	        return XbcExpExpOpKind.S2L_OK;
	    case XbcExpToken.S2R_TT:
	        return XbcExpExpOpKind.S2R_OK;
		}
		return 0;
	}
	
	private XbcExpExpTree readShiftExp() {
		XbcExpExpTree exp = readAddExp();
		int op = translateShiftEt(token);
		while(op > 0) {
			match(token);
			XbcExpExpTree exp1 = new XbcExpExpTree(es.getLinenum(), op);
			exp1.lchild = exp;
			exp1.rchild = readAddExp();
			exp = exp1;
			op = translateShiftEt(token);
		}
		return exp;
	}

	private int translateAddEt(int t) {
		switch(t) {
	    case XbcExpToken.PLS_TT:
	        return XbcExpExpOpKind.PLS_OK;
	    case XbcExpToken.MNS_TT:
	        return XbcExpExpOpKind.MNS_OK;
		}
		return 0;
	}
	
	private XbcExpExpTree readAddExp() {
		XbcExpExpTree exp = readMulExp();
		int op = translateAddEt(token);
		while(op > 0) {
			match(token);
			XbcExpExpTree exp1 = new XbcExpExpTree(es.getLinenum(), op);
			exp1.lchild = exp;
			exp1.rchild = readMulExp();
			exp = exp1;
			op = translateAddEt(token);
		}
		return exp;
	}

	private int translateMulEt(int t) {
		switch(t) {
	    case XbcExpToken.MPT_TT:
	        return XbcExpExpOpKind.MPT_OK;
	    case XbcExpToken.DIV_TT:
	        return XbcExpExpOpKind.DIV_OK;
	    case XbcExpToken.MOD_TT:
	        return XbcExpExpOpKind.MOD_OK;
		}
		return 0;
	}
	
	private XbcExpExpTree readMulExp() {
		XbcExpExpTree exp = readCastExp();
		int op = translateMulEt(token);
		while(op > 0) {
			match(token);
			XbcExpExpTree exp1 = new XbcExpExpTree(es.getLinenum(), op);
			exp1.lchild = exp;
			exp1.rchild = readCastExp();
			exp = exp1;
			op = translateMulEt(token);
		}
		return exp;
	}
	
	private XbcExpExpTree readCastExp() {
		XbcExpExpTree exp = null;
		switch(token) {
		case XbcExpToken.SELFPLS_TT:
			match(token);
			exp = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.PREINC_OK);
			exp.lchild = readCastExp();
			break;
		case XbcExpToken.SELFMNS_TT:
			match(token);
			exp = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.PREDEC_OK);
			exp.lchild = readCastExp();
			break;
		case XbcExpToken.MNS_TT:
			match(token);
			XbcExpExpTree exp1 = readCastExp();
			if(exp1.ek == XbcExpExpKind.CST_EK || (exp1.ok != XbcExpExpVarKind.INT_VK && exp1.ok != XbcExpExpVarKind.DBL_VK)) {
				exp = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.MNS_OK);
				exp.lchild = XbcExpExpTree.newConstIntExptree(es.getLinenum(), "0", expStr);
				exp.rchild = exp1;
			} else {
				if(exp1.ok == XbcExpExpVarKind.INT_VK) {
					exp1.value = 0 - (int)exp1.value;
				} else {
					exp1.value = 0.0 - (double)exp1.value;
				}
			}
			break;
		case XbcExpToken.PLS_TT:
			match(token);
			exp = readCastExp();
			break;
		case XbcExpToken.BNOT_TT:
			match(token);
			exp = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.BNOT_OK);
			exp.lchild = readCastExp();
			break;
		case XbcExpToken.NOT_TT:
			match(token);
			exp = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.NOT_OK);
			exp.lchild = readCastExp();
			break;
		default:
			exp = readPfExp();
			break;
		}
		return exp;
	}

	private boolean isPfEt(int t) {
		switch(t) {
	    case XbcExpToken.PRD_TT:
	    case XbcExpToken.LPRN_TT:
	        return true;
		}
		return false;
	}
	
	private void readFuncPlist(XbcExpExpTree exp, String fname) {
		while(token != XbcExpToken.RPRN_TT) {
			XbcExpExpTree pexp = readAsnExp();
			exp.params.add(pexp);
			if(token == XbcExpToken.CMM_TT) {
				match(token);
				if(token == XbcExpToken.RPRN_TT) {
					throw new XbcExpParseException(es.getLinenum() + "行: 解析函数【" + fname + "】的参数时出错。" + " 公式: " + expStr);
				}
			} else if(token != XbcExpToken.RPRN_TT) {
				throw new XbcExpParseException(es.getLinenum() + "行: 解析函数【" + fname + "】的参数时出错。" + " 公式: " + expStr);
			}
		}
		match(XbcExpToken.RPRN_TT);
	}
	
	private XbcExpExpTree readPfExp() {
		XbcExpExpTree exp = readFactExp();
		if(exp == null) {
			return null;
		}
		if(exp.ek == XbcExpExpKind.CST_EK && exp.ok != XbcExpExpVarKind.STR_VK) {
			return exp;
		}
		while(isPfEt(token)) {
			switch(token) {
			case XbcExpToken.PRD_TT:
				match(token);
				if(token != XbcExpToken.EXPNAME_TT) {
					throw new XbcExpParseException(es.getLinenum() + "行: 点号后面应该接一个成员变量名或方法名。" + " 公式: " + expStr);
				}
				XbcExpExpTree exp2 = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.PRD_OK);
				exp2.lchild = exp;
				exp2.value =  es.getTokenContent();
				match(token);
				exp = exp2;
				break;
			case XbcExpToken.LPRN_TT:
				match(token);
				XbcExpExpTree exp1 = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.FUNC_OK);
				exp1.lchild = exp;
				exp1.params = new ArrayList<>();
				paramSet.remove((String)exp.value);
				es.setFunctionDiv((String)exp.value);
				readFuncPlist(exp1, (String)exp.value);
				exp = exp1;
				break;
			}
		}
		if(token == XbcExpToken.SELFMNS_TT || token == XbcExpToken.SELFPLS_TT) {
			XbcExpExpTree exp1 = null;
			if(token == XbcExpToken.SELFMNS_TT) {
				exp1 = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.AFTDEC_OK);
			} else {
				exp1 = new XbcExpExpTree(es.getLinenum(), XbcExpExpOpKind.AFTINC_OK);
			}
			match(token);
			exp1.lchild = exp;
			exp = exp1;
		}
		return exp;
	}
	
	private XbcExpExpTree readFactExp() {
		XbcExpExpTree exp = null;
		switch(token) {
		case XbcExpToken.EXPNAME_TT:
			String tokenContent = es.getTokenContent();
			if(tokenContent.equals("true")) {
				exp = XbcExpExpTree.newConstBoolExptree(es.getLinenum(), true);
			} else if(tokenContent.equals("false")) {
				exp = XbcExpExpTree.newConstBoolExptree(es.getLinenum(), false);
			} else if(tokenContent.equals("null")) {
				exp = XbcExpExpTree.newConstNullExptree(es.getLinenum());
			} else {
				exp = XbcExpExpTree.newIdExptree(es.getLinenum(), tokenContent);
				if(!removedParamSet.contains(tokenContent)) {
					paramSet.add(tokenContent);
				}
			}
			match(token);
			break;
		case XbcExpToken.INT_TT:
			exp = XbcExpExpTree.newConstIntExptree(es.getLinenum(), es.getTokenContent(), expStr);
			match(token);
			break;
		case XbcExpToken.DBL_TT:
			exp = XbcExpExpTree.newConstDblExptree(es.getLinenum(), es.getTokenContent(), expStr);
			match(token);
			break;
		case XbcExpToken.STR_TT:
			exp = XbcExpExpTree.newConstStrExptree(es.getLinenum(), es.getTokenContent());
			match(token);
			break;
		case XbcExpToken.LPRN_TT:
			match(token);
			exp = readAsnExp();
			match(XbcExpToken.RPRN_TT);
			break;
		case XbcExpToken.NEW_TT:     // new
			exp = readNewExp();
			break;
		case XbcExpToken.INVOKE_TT:
			exp = readInvokeExp();
			break;
		default:
			throw new XbcExpParseException(es.getLinenum() + "行: 读取表达式出错: " + es.getTokenContent() + ", 公式: " + expStr);
		}
		return exp;
	}
}
