package xbc.formula.engine.exp.scanner;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import xbc.formula.engine.exp.datatype.XbcExpToken;
import xbc.formula.engine.exp.exception.XbcExpScanException;
import xbc.formula.engine.exp.vo.XbcExpDrawnInfoVo;

/**
 * 从表达式字符串中提取token
 * 
 * @author leichangchun
 */
public class XbcExpScanner {

	private static final char EOF = '\0';

	private static final int SCAN_START_STATE = 0;      // start-state
	private static final int SCAN_LC_STATE    = 1;      // line-comment-state
	private static final int SCAN_C_STATE     = 2;      // comment-state
	private static final int SCAN_EN_STATE    = 3;      // expname-state
	//private static final int SCAN_EEN1_STATE  = 4;      // e-expname1-state
	private static final int SCAN_EEN_STATE   = 5;      // e-expname-state
	private static final int SCAN_INT1_STATE  = 6;      // integer1-state
	private static final int SCAN_INT_STATE   = 7;      // integer-state
	private static final int SCAN_OCT_STATE   = 8;      // octal-state
	private static final int SCAN_HEX_STATE   = 9;      // hex-state
	private static final int SCAN_DBL_STATE   = 10;     // double-state
	private static final int SCAN_EDBL1_STATE = 11;     // e-double-state
	private static final int SCAN_EDBL_STATE  = 12;     // e-double-state
	private static final int SCAN_STR_STATE   = 13;     // string-state
	private static final int SCAN_DONE_STATE  = 20;     // finished-state

//	private static final int SCANF_INT  = 1;
//	private static final int SCANF_OCT  = 2;
//	private static final int SCANF_HEX  = 3;
//	private static final int SCANF_DBL  = 4;
//	private static final int SCANF_EDBL = 5;

	private String expStr;
	private String oriExpStr;

	private int pos;

	private StringBuilder tokenContent;

	private int token;

	private boolean mkDiv;

	private List<XbcExpDrawnInfoVo> divs;

	private StringBuilder divContent;

	private int linenum;

	public XbcExpScanner(String expStr) {
		this(expStr, false);
	}

	public XbcExpScanner(String expStr, boolean mkDiv) {
		this.oriExpStr = expStr;
		readExpContent(expStr);
		this.mkDiv = mkDiv;
		pos = 0;
		tokenContent = new StringBuilder();
		if(mkDiv) {
			divs = new ArrayList<>();
		}
		linenum = 1;
	}
	
	private void readExpContent(String str) {
		File f = new File(str);
		if(!f.exists() || !f.isFile()) {
			expStr = str;
			return;
		}
		StringBuilder sb = new StringBuilder();
		BufferedReader reader = null;
		try {
			reader = new BufferedReader(new FileReader(str));
			String line = reader.readLine();
			while(line != null) {
				sb.append(line);
				sb.append("\n");
				line = reader.readLine();
			}
			expStr = sb.toString();
		} catch(IOException ix) {
			//ix.printStackTrace();
			expStr = str;
		} finally {
			if(reader != null) {
				try {
					reader.close();
				} catch (IOException e) {
					//e.printStackTrace();
					expStr = str;
				}
			}
		}
	}

	public int getLinenum() {
		return linenum;
	}

	public String getTokenContent() {
		return tokenContent.toString();
	}

	public List<XbcExpDrawnInfoVo> getDrawnInfoVos() {
		return divs;
	}

	public void setFunctionDiv(String funcName) {
		if(funcName == null) {
			return;
		}
		if(divs != null && divs.size() > 0) {
			for(XbcExpDrawnInfoVo div : divs) {
				if(div.getType() == XbcExpDrawnInfoVo.VARIABLE_TYPE && funcName.equals(div.getText())) {
					div.setType(XbcExpDrawnInfoVo.FUNCTION_TYPE);
				}
			}
		}
	}

	private void addDiv(int type) {
		if(!mkDiv) {
			return;
		}
		if(divContent.length() == 0) {
			return;
		}
		int dspos = 0;
		if(divs.size() > 0) {
			dspos = divs.get(divs.size() - 1).getEnd() + 1;
		}
		XbcExpDrawnInfoVo div = new XbcExpDrawnInfoVo();
		div.setType(type);
		div.setStart(dspos);
		int end = dspos + divContent.length() - 1;
		div.setEnd(end);
		div.setText(divContent.toString());
		divs.add(div);
		divContent = new StringBuilder();
	}

	private void appendDivChar(char c) {
		if(mkDiv) {
			divContent.append(c);
		}
	}

	public int readToken() {
		tokenContent = new StringBuilder();
		if(mkDiv) {
			divContent = new StringBuilder();
		}
		char c = EOF, lastC, strC = EOF, nextC;
		int state = SCAN_START_STATE;
		//int nFlag;
		while(state != SCAN_DONE_STATE) {
			lastC = c;
			c = getNextChar();
			switch(state) {
			case SCAN_START_STATE:
				if(isSpaceChar(c)) {
					appendDivChar(c);
					break;
				}
				if(mkDiv && divContent.length() > 0) {
					addDiv(XbcExpDrawnInfoVo.SPACE_TYPE);
				}
				if(Character.isAlphabetic(c) || c == '_') {
					state = SCAN_EN_STATE;
					ungetNextChar();
					break;
				}
	            if(c > '0'&&c <= '9') {
	                tokenContent.append(c);
					appendDivChar(c);
	                state = SCAN_INT_STATE;
	                break;
	            }
	            switch(c) {
	            case '0':
					appendDivChar(c);
	            	state = SCAN_INT1_STATE;
	                break;
	            case '\'':
					appendDivChar(c);
	                strC = '\'';
	                state = SCAN_STR_STATE;
	                break;
	            case '"':
					appendDivChar(c);
	            	strC = '"';
	                state = SCAN_STR_STATE;
	                break;
	            case '#':
					appendDivChar(c);
	                state = SCAN_LC_STATE;
	                break;
	            case '/':
					appendDivChar(c);
	            	nextC = getNextChar();
	                if(nextC == '/') {
						appendDivChar(nextC);
	                    state = SCAN_LC_STATE;
	                }   else if(nextC == '*') {
						appendDivChar(nextC);
	                    lastC = 0;
	                    state  = SCAN_C_STATE;
	                }   else if(nextC == '=') {
						appendDivChar(nextC);
						addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
	                    token = XbcExpToken.DIVASN_TT;
	                    state = SCAN_DONE_STATE;
	                }   else {
						addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
	                	token = XbcExpToken.DIV_TT;
	                    state = SCAN_DONE_STATE;
	                    ungetNextChar();
	                }
	                break;
	            case '-':
					appendDivChar(c);
	            	nextC = getNextChar();
	                if(nextC == '-') {
						appendDivChar(nextC);
	                    token = XbcExpToken.SELFMNS_TT;
	                }   else if(nextC == '=') {
						appendDivChar(nextC);
	                    token = XbcExpToken.MNSASN_TT;
	                }   else {
	                    token = XbcExpToken.MNS_TT;
	                    ungetNextChar();
	                }
					state = SCAN_DONE_STATE;
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
	                break;
	            case '!':
					appendDivChar(c);
	            	nextC = getNextChar();
	                if(nextC == '=') {
						appendDivChar(nextC);
	                    token = XbcExpToken.NEQ_TT;
	                }   else {
	                    token = XbcExpToken.NOT_TT;
	                    ungetNextChar();
	                }
					state = SCAN_DONE_STATE;
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
	                break;
	            case '%':
					appendDivChar(c);
	            	nextC = getNextChar();
	                if(nextC == '=') {
						appendDivChar(nextC);
	                    token = XbcExpToken.MODASN_TT;
	                }   else {
	                    token = XbcExpToken.MOD_TT;
	                    ungetNextChar();
	                }
					state = SCAN_DONE_STATE;
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
	                break;
	            case '&':
					appendDivChar(c);
	            	nextC = getNextChar();
	                if(nextC == '&') {
						appendDivChar(nextC);
	                    token = XbcExpToken.AND_TT;
	                }   else if(nextC == '=') {
						appendDivChar(nextC);
	                    token = XbcExpToken.BANDASN_TT;
	                }   else {
	                    token = XbcExpToken.BAND_TT;
	                    ungetNextChar();
	                }
					state = SCAN_DONE_STATE;
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
	                break;
	            case '(':
					appendDivChar(c);
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
                    token = XbcExpToken.LPRN_TT;
	                state = SCAN_DONE_STATE;
	                break;
	            case ')':
					appendDivChar(c);
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
                    token = XbcExpToken.RPRN_TT;
	                state = SCAN_DONE_STATE;
	                break;
	            case '*':
					appendDivChar(c);
	            	nextC = getNextChar();
	                if(nextC == '=') {
						appendDivChar(nextC);
	                    token = XbcExpToken.MPTASN_TT;
	                }   else {
	                    token = XbcExpToken.MPT_TT;
	                    ungetNextChar();
	                }
					state = SCAN_DONE_STATE;
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
	                break;
	            case ',':
					appendDivChar(c);
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
                    token = XbcExpToken.CMM_TT;
	                state = SCAN_DONE_STATE;
	                break;
	            case '.':
					appendDivChar(c);
	            	nextC = getNextChar();
	            	if(nextC >= '0' && nextC <= '9') {
		            	tokenContent.append('0');
		            	tokenContent.append('.');
		            	tokenContent.append(nextC);
		                state = SCAN_DBL_STATE;
	            	}   else {
	            		ungetNextChar();
	            		token = XbcExpToken.PRD_TT;
		                state = SCAN_DONE_STATE;
	            	}
	            	break;
	            case ':':
					appendDivChar(c);
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
                    token = XbcExpToken.CLN_TT;
	                state = SCAN_DONE_STATE;
	                break;
	            case ';':
					appendDivChar(c);
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
                    token = XbcExpToken.SEMI_TT;
	                state = SCAN_DONE_STATE;
	                break;
	            case '?':
					appendDivChar(c);
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
                    token = XbcExpToken.QRY_TT;
	                state = SCAN_DONE_STATE;
	                break;
	            case '^':
					appendDivChar(c);
	            	nextC = getNextChar();
	                if(nextC == '=') {
						appendDivChar(nextC);
	                    token = XbcExpToken.BXORASN_TT;
	                }   else {
	                    token = XbcExpToken.BXOR_TT;
	                    ungetNextChar();
	                }
					state = SCAN_DONE_STATE;
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
	                break;
	            case '|':
					appendDivChar(c);
	            	nextC = getNextChar();
	                if(nextC == '|') {
						appendDivChar(nextC);
	                    token = XbcExpToken.OR_TT;
	                }   else if(nextC == '=') {
						appendDivChar(nextC);
	                    token = XbcExpToken.BORASN_TT;
	                }   else {
	                    token = XbcExpToken.BOR_TT;
	                    ungetNextChar();
	                }
					state = SCAN_DONE_STATE;
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
	                break;
	            case '~':
					appendDivChar(c);
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
                    token = XbcExpToken.BNOT_TT;
	                state = SCAN_DONE_STATE;
	                break;
	            case '+':
					appendDivChar(c);
	            	nextC = getNextChar();
	                if(nextC == '+') {
						appendDivChar(nextC);
	                    token = XbcExpToken.SELFPLS_TT;
	                }   else if(nextC == '=') {
						appendDivChar(nextC);
	                    token = XbcExpToken.PLSASN_TT;
	                }   else {
	                    token = XbcExpToken.PLS_TT;
	                    ungetNextChar();
	                }
					state = SCAN_DONE_STATE;
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
	                break;
	            case '<':
					appendDivChar(c);
	            	nextC = getNextChar();
	                if(nextC == '<') {
						appendDivChar(nextC);
	                    c = getNextChar();
	                    if(c == '=') {
							appendDivChar(c);
		                    token = XbcExpToken.S2LASN_TT;
	                    } else {
		                    token = XbcExpToken.S2L_TT;
	                        ungetNextChar();
	                    }
	                }   else if(nextC == '=') {
						appendDivChar(nextC);
	                    token = XbcExpToken.LEQ_TT;
	                }   else {
	                    token = XbcExpToken.LT_TT;
	                    ungetNextChar();
	                }
					state = SCAN_DONE_STATE;
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
	                break;
	            case '=':
					appendDivChar(c);
	            	nextC = getNextChar();
	                if(nextC == '=') {
						appendDivChar(nextC);
	                    token = XbcExpToken.EQ_TT;
	                }   else {
	                    token = XbcExpToken.ASN_TT;
	                    ungetNextChar();
	                }
					state = SCAN_DONE_STATE;
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
	                break;
	            case '>':
					appendDivChar(c);
	            	nextC = getNextChar();
	                if(nextC == '>') {
						appendDivChar(nextC);
	                    c = getNextChar();
	                    if(c == '=') {
							appendDivChar(c);
		                    token = XbcExpToken.S2RASN_TT;
	                    } else {
		                    token = XbcExpToken.S2R_TT;
	                        ungetNextChar();
	                    }
	                }   else if(nextC == '=') {
						appendDivChar(nextC);
	                    token = XbcExpToken.GEQ_TT;
	                }   else {
	                    token = XbcExpToken.GT_TT;
	                    ungetNextChar();
	                }
					state = SCAN_DONE_STATE;
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
	                break;
	            case '{':
					appendDivChar(c);
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
	            	token = XbcExpToken.LSQB_TT;
					state = SCAN_DONE_STATE;
					break;
	            case '}':
					appendDivChar(c);
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
	            	token = XbcExpToken.RSQB_TT;
					state = SCAN_DONE_STATE;
					break;
	            case '[':
					appendDivChar(c);
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
	            	token = XbcExpToken.LBKT_TT;
					state = SCAN_DONE_STATE;
					break;
	            case ']':
					appendDivChar(c);
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
					token = XbcExpToken.RBKT_TT;
					state = SCAN_DONE_STATE;
					break;
	            case '@':
					appendDivChar(c);
					addDiv(XbcExpDrawnInfoVo.OPERATOR_TYPE);
					token = XbcExpToken.AT_TT;
					state = SCAN_DONE_STATE;
					break;
	            case EOF:
                    token = XbcExpToken.EOF_TT;
	                state = SCAN_DONE_STATE;
	                break;
	            default:
	            	throw new XbcExpScanException(linenum + "行: " + c + " 无法识别的符号。" + " 公式: " + oriExpStr);
	            }
				break;
			case SCAN_LC_STATE:    // line-comment-state
				if(c == EOF) {
					addDiv(XbcExpDrawnInfoVo.COMMENT_TYPE);
					token = XbcExpToken.EOF_TT;
	                state = SCAN_DONE_STATE;
				} else if(c == 10) {
					appendDivChar(c);
					state = SCAN_START_STATE;
				}
				break;
			case SCAN_C_STATE:     // comment-state
				appendDivChar(c);
				if(c == EOF) {
					throw new XbcExpScanException(linenum + "行: 【/*】注释没有结束符【*/】。" + " 公式: " + oriExpStr);
				}   else if(lastC == '*'&&c == '/') {
					addDiv(XbcExpDrawnInfoVo.COMMENT_TYPE);
	                state = SCAN_START_STATE;
	            }
				break;
			case SCAN_EN_STATE:    // expname-state
	            if(Character.isAlphabetic(c) || c == '_' || (c >= '0' && c <= '9')) {
					appendDivChar(c);
	            	tokenContent.append(c);
	            } else {
					addDiv(XbcExpDrawnInfoVo.VARIABLE_TYPE);
					token = XbcExpToken.EXPNAME_TT;
                    state = SCAN_DONE_STATE;
                    ungetNextChar();
	            }
	            break;
	        case SCAN_EEN_STATE:   // e-expname-state
	            if(c >= '0' && c <= '9') {
					appendDivChar(c);
	            	tokenContent.append(c);
	            } else {
					addDiv(XbcExpDrawnInfoVo.VARIABLE_TYPE);
					token = XbcExpToken.EXPNAME_TT;
                    state = SCAN_DONE_STATE;
                    ungetNextChar();
	            }
	            break;
	        case SCAN_INT1_STATE:  // integer1-state
	            if(c == 'x'||c == 'X') {
					appendDivChar(c);
	            	state = SCAN_HEX_STATE;
				} else if(c >= '0'&&c <= '7') {
					appendDivChar(c);
	            	tokenContent.append(c);
	                state = SCAN_OCT_STATE;
	            }   else if(c == '.') {
					appendDivChar(c);
	            	tokenContent.append('0');
	            	tokenContent.append('.');
	                state = SCAN_DBL_STATE;
	            }   else {
	                //nFlag = SCANF_INT;
					addDiv(XbcExpDrawnInfoVo.CONSTANT_NUMERIC_TYPE);
					token = XbcExpToken.INT_TT;
	                state = SCAN_DONE_STATE;
	            	tokenContent.append('0');
                    ungetNextChar();
	            }
	            break;
	        case SCAN_INT_STATE:   // integer-state
	            if(c >= '0' && c <= '9') {
					appendDivChar(c);
	            	tokenContent.append(c);
	            } else if(c == '.') {
					appendDivChar(c);
	            	tokenContent.append(c);
	                state = SCAN_DBL_STATE;
	            }   else if(c == 'e'||c == 'E') {
					appendDivChar(c);
	            	tokenContent.append('E');
	                state = SCAN_EDBL1_STATE;
	            }   else if(Character.isAlphabetic(c) || c == '"') {
	            	throw new XbcExpScanException(linenum + "行: 整数【" + tokenContent.toString() + "】后面不能接字符: " + c + ", 公式: " + oriExpStr);
	            }   else {
	                //nFlag = SCANF_INT;
					addDiv(XbcExpDrawnInfoVo.CONSTANT_NUMERIC_TYPE);
					token = XbcExpToken.INT_TT;
	                state = SCAN_DONE_STATE;
                    ungetNextChar();
	            }
	            break;
	        case SCAN_OCT_STATE:   // octal-state
	            if(c >= '0'&&c <= '7') {
					appendDivChar(c);
	            	tokenContent.append(c);
	            } else if(Character.isAlphabetic(c)||c == '8'||c == '9'||c == '"') {
	            	throw new XbcExpScanException(linenum + "行: 八进制整数【" + tokenContent.toString() + "】后面不能接字符: " + c + ", 公式: " + oriExpStr);
	            }   else {
	                //nFlag = SCANF_OCT;
					addDiv(XbcExpDrawnInfoVo.CONSTANT_NUMERIC_TYPE);
					token = XbcExpToken.INT_TT;
	                state = SCAN_DONE_STATE;
                    ungetNextChar();
	            }
	            break;
	        case SCAN_HEX_STATE:   // hex-state
	            if((c >= '0' && c <= '9') || (c >= 'a' && c <= 'f') || (c >= 'A' && c <= 'F')) {
					appendDivChar(c);
	            	tokenContent.append(c);
	            } else if(Character.isAlphabetic(c) ||c == '"') {
	            	throw new XbcExpScanException(linenum + "行: 十六进制整数【" + tokenContent.toString() + "】后面不能接字符: " + c + ", 公式: " + oriExpStr);
	            }   else {
	                //nFlag = SCANF_HEX;
					addDiv(XbcExpDrawnInfoVo.CONSTANT_NUMERIC_TYPE);
					token = XbcExpToken.INT_TT;
	                state = SCAN_DONE_STATE;
                    ungetNextChar();
	            }
	            break;
	        case SCAN_DBL_STATE:   // double-state
	            if(c >= '0' && c <= '9') {
					appendDivChar(c);
	            	tokenContent.append(c);
	            } else if(c == 'e'||c == 'E') {
					appendDivChar(c);
	            	tokenContent.append('E');
	                state = SCAN_EDBL1_STATE;
	            } else if(Character.isAlphabetic(c) ||c == '"') {
	            	throw new XbcExpScanException(linenum + "行: 浮点数【" + tokenContent.toString() + "】后面不能接字符: " + c + ", 公式: " + oriExpStr);
	            }   else {
	                //nFlag = SCANF_DBL;
					addDiv(XbcExpDrawnInfoVo.CONSTANT_NUMERIC_TYPE);
					token = XbcExpToken.DBL_TT;
	                state = SCAN_DONE_STATE;
                    ungetNextChar();
	            }
	            break;
	        case SCAN_EDBL1_STATE:  // e-double-state
				appendDivChar(c);
	        	tokenContent.append(c);
	            if(c == '+'||c == '-') {
	                c = getNextChar();
					appendDivChar(c);
	                tokenContent.append(c);
	                if(c >= '0' && c <= '9') {
	                    state = SCAN_EDBL_STATE;
	                }   else {
	                	throw new XbcExpScanException(linenum + "行: +-符号后面应该跟数字" + ", 公式: " + oriExpStr);
	                }
	            }   else if(c >= '0' && c <= '9') {
	                state = SCAN_EDBL_STATE;
	            }   else {
	            	throw new XbcExpScanException(linenum + "行: 科学计数法常数不正确。" + ", 公式: " + oriExpStr);
	            }
	            break;
	        case SCAN_EDBL_STATE:  // e-double-state
	            if(c >= '0' && c <= '9') {
					appendDivChar(c);
	            	tokenContent.append(c);
	            } else {
	                //nFlag = SCANF_EDBL;
					addDiv(XbcExpDrawnInfoVo.CONSTANT_NUMERIC_TYPE);
					token = XbcExpToken.DBL_TT;
	                state = SCAN_DONE_STATE;
                    ungetNextChar();
	            }
	            break;
	        case SCAN_STR_STATE:   // string-state
	            if(c == EOF||c == 13||c == 10||c == 12) {
	            	throw new XbcExpScanException(linenum + "行: 字符串常数没有结束【" + tokenContent.toString() + "】" + ", 公式: " + oriExpStr);
	            }   else if(c == strC) {
					appendDivChar(c);
					addDiv(XbcExpDrawnInfoVo.CONSTANT_STRING_TYPE);
					token = XbcExpToken.STR_TT;
	                state = SCAN_DONE_STATE;
	            }   else if(c == 92) {
	            	charEscape();
	            }   else {
					appendDivChar(c);
	            	tokenContent.append(c);
	            }
	            break;
			}
		}
		if(token == XbcExpToken.EXPNAME_TT) {
			String ts = tokenContent.toString();
			if(ts.equals("for")) {
				token = XbcExpToken.FOR_TT;
			} else if(ts.equals("break")) {
				token = XbcExpToken.BREAK_TT;
			} else if(ts.equals("continue")) {
				token = XbcExpToken.CONTINUE_TT;
			} else if(ts.equals("while")) {
				token = XbcExpToken.WHILE_TT;
			} else if(ts.equals("if")) {
				token = XbcExpToken.IF_TT;
			} else if(ts.equals("else")) {
				token = XbcExpToken.ELSE_TT;
			} else if(ts.equals("div0")) {
				token = XbcExpToken.DIV0_TT;
			} else if(ts.equals("exit")) {
				token = XbcExpToken.EXIT_TT;
			} else if(ts.equals("new") || ts.equals("create")) {
				token = XbcExpToken.NEW_TT;
			} else if(ts.equals("invoke")) {
				token = XbcExpToken.INVOKE_TT;
			}
		}
		return token;
	}

	private void charEscape() {
	    char c = getNextChar();
	    switch(c) {
	    case 92:
	    	tokenContent.append(c);
	        return;
	    case 'b':
	    	tokenContent.append("\b");
	        return;
	    case 'f':
	    	tokenContent.append("\f");
	        return;
	    case 'n':
	    	tokenContent.append("\n");
	        return;
	    case 'r':
	    	tokenContent.append("\r");
	        return;
	    case 't':
	    	tokenContent.append("\t");
	        return;
	    case '\'':
	    	tokenContent.append("\\");
	        return;
	    case '"':
	    	tokenContent.append("\"");
	        return;
	    case '?':
	    	tokenContent.append("?");
	        return;
	    }
    	throw new XbcExpScanException(linenum + "行: 未知的转义符【" + c + "】" + ", 公式: " + oriExpStr);
	}

	private void ungetNextChar() {
		if(pos <= expStr.length()) {
			char lastC = expStr.charAt(pos - 1);
			if(lastC == 10 || lastC == 12) {
				linenum --;
			}
		}
		pos --;
	}

	private char getNextChar() {
		if(pos >= expStr.length()) {
			pos ++;
			return EOF;
		}
		char c = expStr.charAt(pos);
		if(c == 10 || c == 12) {
			linenum ++;
		}
		pos ++;
		return c;
	}
	
	private boolean isSpaceChar(char c) {
		if(c == ' '||c == 10||c == 13||c == 12||c == 9) {
			return true;
		}
		return false;
	}
}
