package xbc.formula.engine.fml;

import java.lang.reflect.Field;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import xbc.formula.engine.exp.XbcExpCalculator;
import xbc.formula.engine.exp.exception.XbcExpParamException;
import xbc.formula.engine.exp.ext.XbcExpExtFuncManager;
import xbc.formula.engine.exp.vo.XbcExpRuleVo;
import xbc.formula.engine.fml.vo.XbcFmlFormulaIndicatorVo;
import xbc.formula.engine.fml.vo.XbcFmlFormulaVo;

/**
 * 公式计算类
 * 
 * @author leichangchun
 */
public class XbcFmlCalculator {

	/**
	 * 获取公式里面包含的指标的名称列表。
	 * 
	 * @param expStr 公式的表达式
	 * @return 指标名称列表
	 */
	public static List<String> getIndicators(String expStr) {
		List<String> ids = XbcExpCalculator.getParamters(expStr);
		return ids;
	}

	/**
	 * 执行公式的计算
	 *  没有需要脚本计算出指标值的情况，请使用这个方法
	 * 
	 * @param fv 公式
	 * @param params 传递给公式中包含的脚本类指标的脚本方法的参数，或Map类指标的取值源
	 * @return 公式的计算结果
	 */
	public static Object calculte(XbcFmlFormulaVo fv, Object... formulaParams) {
		return calculte(fv, null, formulaParams);
	}
	
	public static Object calculte(XbcFmlFormulaVo fv, List<XbcExpRuleVo> ruleVos, Object... formulaParams) {
		Set<Long> formulaIdSet = new HashSet<>(); // 防止公式循环调用
		Map<String, Object> resultMap = new HashMap<>();
		Object ov = calculte(fv, formulaIdSet, resultMap, ruleVos, formulaParams);
		return ov;
	}
	
	private static Object calculte(XbcFmlFormulaVo fv, Set<Long> formulaIdSet, Map<String, Object> resultMap, List<XbcExpRuleVo> ruleVos, Object[] formulaParams) {
		if(fv == null) {
			throw new XbcExpParamException("需要计算的公式是空值。");
		}
		if(resultMap.containsKey(fv.getFormulaName())) {
			return resultMap.get(fv.getFormulaName());
		}
		Long formulaId = fv.getKeyId();
		if(formulaIdSet.contains(formulaId)) {
			throw new XbcExpParamException("公式不能循环调用，公式id: " + formulaId + ", 公式名称: " + fv.getFormulaName());
		}
		formulaIdSet.add(formulaId);
		Object ret = executeExp(fv, formulaIdSet, resultMap, ruleVos, formulaParams);
		fv.setResult(ret);
		formulaIdSet.remove(formulaId);
		resultMap.put(fv.getFormulaName(), ret);
		return ret;
	}
	
	private static Object executeExp(XbcFmlFormulaVo fv, Set<Long> formulaIdSet, Map<String, Object> resultMap, List<XbcExpRuleVo> ruleVos, Object[] formulaParams) {
		Map<String, Object> pm = new HashMap<>();
		List<XbcFmlFormulaIndicatorVo> ivs = fv.getIndicatorVos();
		if(ivs != null && ivs.size() > 0) {
			for(XbcFmlFormulaIndicatorVo iv : ivs) {
				String key = iv.getIndicatorName();
				Object val = null;
				int qt = iv.getIndicatorType();
				if(qt == XbcFmlFormulaIndicatorVo.SPECIFIED_TYPE) {
					try {
						val = new BigDecimal(iv.getSpecifiedVal());
					} catch(NumberFormatException nx) {
						val = iv.getSpecifiedVal();
					}
				} else if(qt == XbcFmlFormulaIndicatorVo.SCRIPT_TYPE) {
					val = iv.getResult();
				} else if(qt == XbcFmlFormulaIndicatorVo.FORMULA_TYPE) {
					val = calculte(iv.getFormulaVo(), formulaIdSet, resultMap, ruleVos, formulaParams);
					iv.setResult(val);
				} else if(qt == XbcFmlFormulaIndicatorVo.PARAM_TYPE) {
					val = getValueFromParams(iv, formulaParams);
					iv.setResult(val);
				} else if(qt == XbcFmlFormulaIndicatorVo.FUNC_TYPE) {
					val = getValueFromFunc(iv, fv.getFormulaValue(), formulaParams);
					iv.setResult(val);
				} else {
					throw new XbcExpParamException("不支持的指标类型: " + qt + ", 指标keyid: " + iv.getKeyId() + ", 指标名称: " + iv.getIndicatorName());
				}
				pm.put(key, val);
			}
		}
		String expStr = fv.getFormulaValue();
		Object retObj = XbcExpCalculator.calculate(expStr, pm, ruleVos, formulaParams);
		return retObj;
	}
	
	private static Object getValueFromFunc(XbcFmlFormulaIndicatorVo iv, String expStr, Object[] formulaParams) {
		String funcName = iv.getSpecifiedVal();
		List<Object> pvs = new ArrayList<>();
		pvs.add(iv);
		Object ov = XbcExpExtFuncManager.executeFunc(funcName, pvs, expStr, formulaParams);
		return ov;
	}

	@SuppressWarnings("rawtypes")
	private static Object getValueFromParams(XbcFmlFormulaIndicatorVo iv, Object[] formulaParams) {
		if(formulaParams == null) {
			return null;
		}
		// 查找传入的参数里面，是否有Map
		Map map = null;
		for(Object param : formulaParams) {
			if(param != null && Map.class.isAssignableFrom(param.getClass())) {
				map = (Map)param;
				break;
			}
		}
		if(map != null) {
			if(map.containsKey(iv.getSpecifiedVal())) {
				return map.get(iv.getSpecifiedVal());
			}
		}
		// 查找传入的参数里面，是否有包含 iv.getScriptFieldId() 的对象
		Field field = null;
		Object inst = null;
		for(Object param : formulaParams) {
			if(param != null) {
				field = getFieldByName(param.getClass(), iv.getSpecifiedVal());
				if(field != null) {
					inst = param;
					break;
				}
			}
		}
		if(field != null) {
			field.setAccessible(true);
			try {
				return field.get(inst);
			} catch (IllegalArgumentException | IllegalAccessException e) {
				throw new XbcExpParamException(e, "从参数的对象[" + inst.getClass() + "]获取字段[" + iv.getSpecifiedVal() + "]时出错。指标keyid: " + iv.getKeyId());
			}
		}
		throw new XbcExpParamException("传入的参数中没有找到指标(参数名: " + iv.getSpecifiedVal() + ")需要的值。指标keyid: " + iv.getKeyId() + ", 指标名称: " + iv.getIndicatorName());
	}

	private static Field getFieldByName(Class<?> clazz, String name) {
		Field field = null;
		try {
			field = clazz.getDeclaredField(name);
		} catch (NoSuchFieldException | SecurityException e) {
			//
		}
		return field;
	}
}
